import sys
import math
import operator
import re
import heapq
import random
import timeit


# This class holds the tuple (doc_id, position)
# Overloaded operator __gt__ (>), __lt__ (<), __ge__(>=), __le__(<=) __str__
# These operators allows to compare the position
class Position:

    def __init__(self, doc_id, position):
        self.doc_id = doc_id
        self.pos = position

    def __gt__(self, other):
        if self.doc_id > other.doc_id:
            return True
        elif self.doc_id == other.doc_id and self.pos > other.pos:
            return True
        return False

    def __lt__(self, other):
        if self.doc_id < other.doc_id:
            return True
        elif self.doc_id == other.doc_id and self.pos < other.pos:
            return True
        return False

    def __ge__(self, other):
        if self.doc_id > other.doc_id:
            return True
        elif self.doc_id == other.doc_id and (self.pos > other.pos or self.pos == other.pos):
            return True
        return False

    def __le__(self, other):
        if self.doc_id < other.doc_id:
            return True
        elif self.doc_id == other.doc_id and (self.pos < other.pos or self.pos == other.pos):
            return True
        return False

    def __str__(self):
        return str(self.doc_id)+","+str(self.pos)


class InvertedIndex:

    inverted_index = {}         # holds inverted index [term, [document_id, term_frequency, [list of positions]]]
    term_frequency = {}         # holds frequency of every term of the corpus
    document_frequency = {}     # holds document frequency of every term
    document_length = {}        # holds length of every document (length = number of terms in the document)
    total_corpus_term = 0       # total number of terms in the entire corpus
    document_average_length = 0
    number_of_document = 0
    last_index_position = {}
    temp_inverted_index = {}

    def build_inverted_index(self, filename):
        document_number = 0
        filename = open(filename, 'r')
        for document in filename.read().split("\n\n"):
            document_number += 1
            position = 0
            for term in document.split():
                position += 1
                term = re.sub('[^A-Za-z0-9]+', '', term)
                term = term.lower()  # convert every term to lower case
                if term not in self.inverted_index:
                    self.inverted_index[term] = [[document_number, 1, [position]]]
                    self.term_frequency[term] = 1
                    self.document_frequency[term] = 1
                else:
                    self.term_frequency[term] += 1
                    postings_lists = self.inverted_index[term]
                    if document_number == postings_lists[-1][0]:
                        postings_lists[-1][2].append(position)
                        postings_lists[-1][1] += 1
                    else:
                        postings_lists += [[document_number, 1, [position]]]
                        self.document_frequency[term] += 1
            self.document_length[document_number] = position
            self.total_corpus_term += position
        filename.close()

        # Sorting the inverted_index based on the term
        for key in sorted(self.inverted_index.keys()):
            self.temp_inverted_index[key] = self.inverted_index[key]
        self.inverted_index = self.temp_inverted_index
        self.number_of_document = document_number
        self.document_average_length = self.total_corpus_term / self.number_of_document
        # #print("Average Length ", self.document_average_length)

    # Helper method to return list of documents the given term appears in
    def get_documents_by_term(self, term):
        documents = []
        if self.is_present(term):
            for posting_list in self.inverted_index[term]:
                documents.append(posting_list[0])
        return documents

    # Helper method to return term_frequency within the document
    def get_document_frequency(self, term, docid):
        if self.is_present(term):
            for posting_list in self.inverted_index[term]:
                if docid == posting_list[0]:
                    return posting_list[1]
        return 0  # return 0 if the term is not in the vocabulary

    # Helper method to return if the term is present in the corpus
    def is_present(self, term):
        if term in self.inverted_index:
            return True
        return False

    # helper method to find the relative position based on the cursor
    # return the Position object after locating the cursor
    def find_position_for_cursor(self, term, cursor):
        posting_list = self.inverted_index[term]
        for posting in posting_list:
            position_list = posting[2]
            if cursor < len(position_list):
                # return when the cursor is appropriate index of the posting list
                return Position(posting[0], position_list[cursor])
            else:
                cursor -= len(position_list)
        return None

    def first(self, term):
        if self.is_present(term):
            return Position(self.first_doc(term), self.inverted_index[term][0][2][0])
        return None

    def last(self, term):
        if self.is_present(term):
            return Position(self.last_doc(term), self.inverted_index[term][-1][2][-1])
        return None

    def first_doc(self, term):
        if self.is_present(term):
            return self.inverted_index[term][0][0]
        return None

    def last_doc(self, term):
        if self.is_present(term):
            return self.inverted_index[term][-1][0]
        return None

    # return the previous document in which the term is present
    def prev_doc(self, term, current):
        if self.is_present(term):
            prev_pos = self.prev(term, current)
            if prev_pos is None:
                return None
            else:
                # Keep finding the prev term until the doc_id of the prev term is less than current doc_id
                while prev_pos is not None and prev_pos.doc_id >= current.doc_id:
                    prev_pos = self.prev(term, prev_pos)
                if prev_pos is not None:
                    return prev_pos.doc_id
        return None

    def next_doc(self, term, current):
        if self.is_present(term):
            next_pos = self.next(term, current)
            if next_pos is None:
                return None
            else:
                # Keep finding the next term until the doc_id of the prev term is greater than current doc_id
                while next_pos is not None and next_pos.doc_id <= current.doc_id:
                    next_pos = self.next(term, next_pos)
                if next_pos is not None:
                    return next_pos.doc_id
        return None

    # Finds the next occurrence of the term after the current position
    def next(self, term, current):
        # #print(current)
        last_position = self.last(term)
        first_position = self.first(term)
        low = 0
        if last_position is None or last_position <= current:  # If current position is after the last occurrence
            return None
        elif first_position > current:   # If current position is before the first occurrence return first position
            self.last_index_position[term] = 0
            return first_position
        else:
            if term in self.last_index_position\
                    and self.last_index_position[term] > 0:
                last_static_position = self.find_position_for_cursor(term, self.last_index_position[term]-1)
                if last_static_position is not None and last_static_position <= current:
                    # If the last position is present we begin the search from there
                    low = self.last_index_position[term] - 1

        jump = 1
        high = low + jump
        high_pos = self.find_position_for_cursor(term, high)

        while high < int(self.term_frequency[term]) \
                and high_pos is not None \
                and high_pos <= current:
            low = high
            jump = 2 * jump
            high = low + jump
            high_pos = self.find_position_for_cursor(term, high)  # Find the relative position of high cursor

        if high >= int(self.term_frequency[term]):
            high = int(self.term_frequency[term]) - 1

        # Performs binary search on the low and high range
        low, high = self.binary_search(term, low, high, current, operator.lt)
        self.last_index_position[term] = high
        return self.find_position_for_cursor(term, self.last_index_position[term])

    # Finds the previous occurrence of the term after the current position
    def prev(self, term, current):
        # #print(current)
        last_position = self.last(term)
        first_position = self.first(term)
        high = self.term_frequency[term] - 1
        if first_position is None or current <= first_position:
            return None
        elif last_position < current:
            self.last_index_position[term] = self.term_frequency[term] - 1
            return last_position
        else:
            if term in self.last_index_position\
                    and self.last_index_position[term] < high:
                last_static_position = self.find_position_for_cursor(term, self.last_index_position[term] + 1)
                if last_static_position is not None and last_static_position >= current:
                    high = self.last_index_position[term] + 1

        jump = 1
        low = high - jump

        if low >= 0:
            low_pos = self.find_position_for_cursor(term, low)

        while low >= 0 \
                and low_pos is not None \
                and low_pos >= current:
            high = low
            jump = 2 * jump
            low = high - jump

            if low >= 0:
                low_pos = self.find_position_for_cursor(term, low)

        if low < 0:
            low = 0

        low, high = self.binary_search(term, low, high, current, operator.lt)
        self.last_index_position[term] = low
        return self.find_position_for_cursor(term, self.last_index_position[term])

    def binary_search(self, term, b_low, b_high, current, op):
        while b_high - b_low > 1:
            mid = b_low + math.floor((b_high-b_low)/2)

            mid_pos = self.find_position_for_cursor(term, mid)

            if op(mid_pos, current):
                b_low = mid
            else:
                b_high = mid

        return b_low, b_high

    def find_BM25_rank_without_heap(self, doc_list, query_terms, k):
        results = []
        for doc in doc_list:
            score = 0
            for term in query_terms:
                d = self.get_documents_by_term(term)
                if doc in d:  # document does not have the term
                    tf = self.TF_BM25(term, doc)
                    idf = self.IDF_BM25(term)
                    s = tf * idf
                    score = score + s
            if len(results) < k:
                heapq.heappush(results, (score, doc))
            else:
                heapq.heappushpop(results, (score, doc))
        results.sort(reverse=True)
        return results

    def find_BM25_rank_with_heap(self, query_terms, max_scores, k):
        results = []                                        # list for the top k document scores (score, doc_id)
        terms = []                                          # term heap contains tuple (doc_id, term)
        excluded_terms = set()
        for i in range(0, k):
            heapq.heappush(results, (0, 0.0))               # Making min-heap of size k

        # Creating term heap by increasing order of next_doc for the term
        for term in query_terms:
            heapq.heappush(terms, (self.first_doc(term), term))

        heapq.heapify(terms)

        t = heapq.nsmallest(1, terms)
        t = t[0]

        # Here self.number_of_documents+1 serves the purpose of +infinity
        while t[0] != self.number_of_document+1:            # loop terminates when all the terms from heap are processed
            d = t[0]
            score = 0
            while t[0] == d:
                score = score + self.TF_BM25(t[1], d) * self.IDF_BM25(t[1])
                next_doc = self.next_doc(t[1], Position(d, 0))
                # When there does not exist next_doc for the term, we add number_of_document+1 to indicate +infinity
                if next_doc is None:
                    heapq.heapreplace(terms, (self.number_of_document+1, t[1]))
                else:
                    heapq.heapreplace(terms, (next_doc, t[1]))
                heapq.heapify(terms)

                t = heapq.nsmallest(1, terms)
                t = t[0]

            # Process the excluded terms, because the scores are valid for documents having non-excluded terms
            for excluded_term in excluded_terms:
                doc = self.next_doc(excluded_term, Position(d - 1, 0))
                if doc is not None and doc == d:
                    score = score + self.TF_BM25(excluded_term, d) * self.IDF_BM25(excluded_term)

            # Get the smallest score from the heap, so that it can be compared with the max_scores list
            smallest = heapq.nsmallest(1, results)
            smallest_score = smallest[0][0]

            # We only add the score if it greater than the lowest score from the heap
            if score > smallest[0][0]:
                heapq.heapreplace(results, (score, d))
                heapq.heapify(results)

            # This loop checks if the smallest score from heap exceeds the max_score for query term
            # max_scores is list containing (max_score, term) for each query term in increasing order of max_score
            for max_score in max_scores:
                if smallest_score > max_score[1]:
                    terms = [t for t in terms if t[1] != max_score[0]]        # remove term from the term heap
                    heapq.heapify(terms)
                    excluded_terms.add(max_score[0])                          # add term to excluded terms
                    smallest_score = smallest_score - max_score[1]

            t = heapq.nsmallest(1, terms)
            t = t[0]

        results.sort(reverse=True)
        return results

    def TF_BM25(self, term, docid):
        k1 = 1.2
        b = 0.75
        ftd = self.get_document_frequency(term, docid)

        numerator = ftd * (k1 + 1)
        denominator = ftd + (k1 * ((1 - b) + (b*(self.document_length[docid]/self.document_average_length))))

        if denominator != 0:
            return float(numerator/denominator)
        return 0

    def IDF_BM25(self, term):
        Nt = self.get_documents_by_term(term)
        if len(Nt) == 0:
            return 0
        else:
            return math.log(float(self.number_of_document)/float(len(Nt)), 2)

    def calculateMaxScoreForTerms(self, query_terms):
        max_term_scores = []
        for t in query_terms:
            score = 2.2*self.IDF_BM25(t)
            max_term_scores.append((t, score))
        return max_term_scores


ii = InvertedIndex()

arguments = len(sys.argv)
if arguments < 3:
    print("please enter argument in the form <file_name number_of_result(int) query(string)>")
else:
    fileName = sys.argv[1]                              # corpus file
    number_of_results = int(sys.argv[2])                # number of results to be returned
    query = sys.argv[3]                                 # query to be processed
    ii.build_inverted_index(fileName)

    # We consider the query to be a disjunctive query and calculate the documents having them
    # This is only used for without heap logic
    doc_list = []
    for term in query.split(" "):
        d = ii.get_documents_by_term(term)
        doc_list = list(set().union(doc_list, d))

    query_list = query.split(" ")
    query_list[:] = (value for value in query_list if value != "_AND" and value != "_OR")
    query_list = [x.lower() for x in query_list]

    # We also find list of all the unique terms in the query for ranking the document
    query_terms = list(set(query_list))

    for t in query_terms:                               # removing term if not present in corpus
        if not ii.is_present(t):
            query_terms.remove(t)

    max_scores = ii.calculateMaxScoreForTerms(query_terms)
    max_scores.sort(key=lambda tup: tup[1])             # sorts max_scores for terms in increasing order of scores

    start = timeit.default_timer()
    result = ii.find_BM25_rank_with_heap(query_terms, max_scores, number_of_results)
    stop = timeit.default_timer()

    #print("Time : ", stop - start)

    query_id = random.randrange(10)
    i = 0
    # print("query_id iter docno rank sim run_id")
    for r in result:
        if r[0] != 0:
            i = i + 1
            print("1", "0", r[1], i, r[0], "my-test")
